{--------------------------------------------------------------
|
|   Name:      RossNull
|
|   Location:  ROSSNULL.PAS
|
|   Purpose:   Calculates setup for Ross Null test
|              Refernce: Modern Optical Enginneering
|              by Warren J. Smith
|
|   Copyright (c) 1990 by Doug George
|   This program may be distributed and used freely
|   provided the copyright notices are not removed.
|
--------------------------------------------------------------}

uses Graph, Crt, Dos;

{--------------------------------------------------------------
|
|   Global Constants and Variables
|
--------------------------------------------------------------}

const
    LDelta = 1e-4;
    SAResidual = 1e-4;
    mSemiDTol = 0.1;
    Pi = 3.141596535;
    HalfPi = Pi / 2;
    XOffset = 31;
    RadToDeg = 180 / Pi;

var
    mSemiDiameter, mRadius, mConicConst : extended;
    lRadiusI, lRadiusM, lThick, lIndex : extended;
    GraphicsOn : boolean;
    XScale, YScale : integer;

{$F+} procedure EGAVGAProc; external; {$F-}
{$L egavga}
{ Build Command: binobj egavga.bgi egavga egavgaproc }

{$F+} procedure HercProc; external; {$F-}
{$L Herc}
{ Build Command: binobj Herc.bgi Herc Hercproc }

{--------------------------------------------------------------
|
|   Name:      DrawAxes
|   Purpose:   Draws scale axes
|
--------------------------------------------------------------}
procedure DrawAxes;
var
    I : integer;
begin
SetColor ( DarkGray );
Line ( 0, GetMaxY div 3, GetMaxX, GetMaxY div 3 );
Line ( 0, 0, 0, GetMaxY );
for I := -GetMaxX div XScale div 2 to GetMaxX div XScale div 2 do
    begin
    Line ( I * XScale + GetMaxX div 2, GetMaxY div 3 - 4,
           I * XScale + GetMaxX div 2, GetMaxY div 3 + 4 );
    end;
for I := GetMaxY div YScale div 2 to -GetMaxY div YScale div 2 do
    begin
    Line ( 0, I * YScale + GetMaxY div 3,
           4, I * YScale + GetMaxY div 3 );
    end;
end;

{--------------------------------------------------------------
|
|   Name:      DrawOptic
|   Purpose:   Draws an optical element (as a vertical line)
|
--------------------------------------------------------------}
procedure DrawOptic (
        X : extended;
        SemiDia : extended);
var
    XI, Y1, Y2 : longint;
begin
SetColor ( LightGray );
XI := Trunc ( X * XScale );
Y1 := Trunc ( SemiDia * YScale ) + GetMaxY div 3;
Y2 := -Trunc ( SemiDia * YScale ) + GetMaxY div 3;
Line ( XI, Y1, XI, Y2 );
end;

{--------------------------------------------------------------
|
|   Name:      DrawRay
|   Purpose:   Draws a ray
|
--------------------------------------------------------------}
procedure DrawRay ( X, Y : extended );
var
    XI, YI : longint;
begin
XI := Trunc ( X * XScale );
YI := Trunc ( Y * YScale ) + GetMaxY div 3;
LineTo ( XI, YI );
end;

{--------------------------------------------------------------
|
|   Name:      Move
|   Purpose:   Moves the pointer
|
--------------------------------------------------------------}
procedure Move ( X, Y : extended );
var
    XI, YI : longint;
begin
XI := Trunc ( X * XScale );
YI := Trunc ( Y * YScale ) + GetMaxY div 3;
MoveTo ( XI, YI );
end;

{--------------------------------------------------------------
|
|   Name:      ToggleScreen
|   Purpose:   Toggles graphics screens
|
--------------------------------------------------------------}
var
    Toggle : integer;
procedure ToggleScreen;
begin
SetVisualPage ( Toggle );
Toggle := (Toggle + 1) mod 2;
SetActivePage ( Toggle );
end;

{--------------------------------------------------------------
|
|   Name:      InitGraphics
|   Purpose:   Sets up the graphics
|
--------------------------------------------------------------}
procedure InitGraphics;
var
    GraphDriver, GraphMode : integer;
begin
GraphDriver := RegisterBGIDriver ( @EGAVGAProc );
GraphDriver := RegisterBGIDriver ( @HercProc );
GraphDriver := Detect;
DetectGraph ( GraphDriver, GraphMode );
GraphMode := 1;
InitGraph ( GraphDriver, GraphMode, '' );
Move ( 0, 0 );
Toggle := 0;
ToggleScreen;
end;

{--------------------------------------------------------------
|
|   Name:      ArcSin
|   Purpose:   Calculate arcsin function
|
--------------------------------------------------------------}
function ArcSin ( X : extended ) : extended;
begin
if abs ( X ) < 1.0 then
    ArcSin := ArcTan ( X / Sqrt ( 1.0 - Sqr ( X ) ) )
else if X < 0.0 then
    ArcSin := -HalfPi
else
    ArcSin := HalfPi;
end;

{--------------------------------------------------------------
|
|   Name:      Power
|   Purpose:   Raise a variable to a power
|
--------------------------------------------------------------}
function Power ( X, Y : extended ) : extended;
begin
if X < 0 then
    Power := -exp ( Y * ln ( -X ) )
else if X = 0 then
    Power := 1
else
    Power := exp ( Y * ln ( X ) );
end;

{--------------------------------------------------------------
|
|   Name:      AsphericReflectRay
|   Purpose:   Traces the ray through a reflection off a mirror
|              with ellipsoid, paraboloid, or hyperboloid surface
|
--------------------------------------------------------------}

procedure AsphericReflectRay (
    var Q : extended;   { Perpendicular distance of ray to vertex of element }
    var U : extended;   { Angle of input }
    var Y : extended;   { Height of intersection }
        p : extended;   { Conic Constant }
        r : extended ); { Radius of surface }
var
    x, xEst : extended;
    A2, A4, A6, A8 : extended;
    G, m : extended;
    Phi : extended;
    Xdir, Ydir : extended;
begin
r := -r;

{ Calculate constants for this aspheric }
A2 := 1 / (2 * r);
A4 := p / (8 * Power ( r, 3 ));                 { 0 for parabola }
A6 := sqr ( p ) / (16 * Power ( r, 5 ));        { 0 for parabola }
A8 := Power ( p, 3 ) / (25.6 * Power ( r, 7 )); { 0 for parabola }

{ Calculate position of incidence if the mirror was flat }
y := Q;
x := 0;

{ Calculate direction cosines }
Ydir := sin ( U );
Xdir := cos ( U );

repeat
    { Calculate value of x given this y for an aspheric surface }
    xEst := A2 * sqr ( y ) + A4 * Power ( y, 4 ) +
            A6 * Power ( y, 6 ) + A8 * Power ( y, 8 );

    { Calculate gradient to correct answer }
    m := -y * (2 * A2 + 4 * A4 * sqr ( y ) +
              6 * A6 * Power ( y, 4 ) + 8 * A8 * Power ( y, 6 ));
    G := (xEst - x) / (Xdir + Ydir * m);

    { Estimate new value }
    x := G * Xdir + x;
    y := G * Ydir + y;
until abs ( x - xEst ) < 0.0001;

{ We now have x and y coordinates of intercept point with aspheroid }
{ Calculate angle between normal and x-axis, then calculate new U }
Phi := ArcTan ( -y / (r - p * x) );
U := 2 * Phi - U;

Q := cos ( U ) * y;

end;


{--------------------------------------------------------------
|
|   Name:      TraceRay
|   Purpose:   Traces the ray through a surface
|
--------------------------------------------------------------}

procedure TraceRay (
    var Q : extended;   { Perpendicular distance of ray to vertex of element }
    var U : extended;   { Angle of input }
    var Y : extended;   { Height of intersection }
        N : extended;   { Refractive index of input media }
        Np : extended;  { Refractive index of output media }
        r : extended ); { Radius of surface }
var
    I, SinI, Ip, Up, G : extended;
begin

{ Meridional Ray Trace Equations }

{ Calculate angle of incidence }
SinI := Q / r - sin ( U );
I := ArcSin ( SinI );

Y := Q * ( 1 + cos ( U + I ) ) / ( cos ( U ) + cos ( I ) );

{ Law of refraction to calculate outgoing ray angle of incidence }
Ip := ArcSin ( N / Np * SinI );

{ Calculate output angle from geometric additions }
Up := U + I - Ip;

{ Calculate output rayheight using formula which is stable for flat surface }
G := Q / (cos ( U ) + cos ( I ));
Q := G * (cos ( Up ) + cos ( Ip )); { Actually Qp }

{ Copy new Up into U }
U := Up;

end;

{--------------------------------------------------------------
|
|   Name:      Transfer
|   Purpose:   Transfers the ray to the next surface
|
--------------------------------------------------------------}

procedure Transfer (
    var Q  : extended;   { Perpendicular distance of output ray to element vertex }
        Up : extended;   { Angle of output }
        t  : extended ); { Distance to next surface }

begin

{ Transfer to Next Surface }
Q := Q - t * sin ( Up );
end;

{--------------------------------------------------------------
|
|   Name:      CalculateLi
|   Purpose:   Calculates the distance to the image from the
|              lens' last surface
|
--------------------------------------------------------------}

procedure CalculateLi (
        Y  : extended;               { Height of ray at lens' first surface }
        L  : extended;               { Distance of object from lens }
        S  : extended;               { Distance to mirror }
    var Li : extended;               { Distance to image }
    var Y1 : extended;               { Ray height at lens' second surface }
    var Y2 : extended );             { Ray height at mirror }

var
    Q, U : extended;

begin
if GraphicsOn then ClearViewPort;
if GraphicsOn then DrawAxes;
if GraphicsOn then DrawOptic ( L, mSemiDiameter * 0.25 );
if GraphicsOn then DrawOptic ( L + lThick, mSemiDiameter * 0.25 );
if GraphicsOn then DrawOptic ( L + lThick + S, mSemiDiameter );
if GraphicsOn then Move ( 0, 0 );

{ Opening: Calculate initial U and Q }
U := arctan ( Y / -L );
Q := -L * sin ( U );

{ Trace through each element }
TraceRay ( Q, U, Y, 1.0, lIndex, lRadiusI );
if GraphicsOn then SetColor ( Green );
if GraphicsOn then DrawRay ( L, Y );

Transfer ( Q, U, lThick );
TraceRay ( Q, U, Y1, lIndex, 1.0, lRadiusM );
if GraphicsOn then SetColor ( Yellow );
if GraphicsOn then DrawRay ( L + lThick, Y1 );

Transfer   ( Q, U, S );
AsphericReflectRay ( Q, U, Y2, mConicConst, mRadius );
if GraphicsOn then SetColor ( Red );
if GraphicsOn then DrawRay ( L + lThick + S, Y2 );

Transfer ( Q, U, -S );
TraceRay ( Q, U, Y, 1.0, lIndex, lRadiusM );
if GraphicsOn then SetColor ( Red );
if GraphicsOn then DrawRay ( L + lThick, Y );

Transfer ( Q, U, -lThick );
TraceRay ( Q, U, Y, lIndex, 1.0, lRadiusI );
if GraphicsOn then SetColor ( Yellow );
if GraphicsOn then DrawRay ( L, Y );

{ Closing: calculate distance }
Li := Q / sin ( U );

if GraphicsOn then SetColor ( Green );
if GraphicsOn then DrawRay ( L + Li, 0 );
if GraphicsOn then ToggleScreen;
end;

{--------------------------------------------------------------
|
|   Name:      CalculateLo
|   Purpose:   Calculates the required distance from the
|              object to the lens' first surface
|
--------------------------------------------------------------}

procedure CalculateLo (
        Y      : extended;           { Height of ray at lens }
        S      : extended;           { Distance from Lens to Mirror }
    var Lo     : extended );         { Calculated Lo }
var
    Li, Y1, Y2 : extended;
begin
CalculateLi ( Y, Lo, S, Li, Y1, Y2 );
repeat
    { If not close enough, re-iterate }
    if abs (Lo + Li) >= LDelta then
        begin
        Lo := Lo - (Lo + Li) * 0.3333333333;
        end;

    CalculateLi ( Y, Lo, S, Li, Y1, Y2 );

until abs (Lo + Li) < LDelta;
end;

{--------------------------------------------------------------
|
|   Name:      CalculateSA
|   Purpose:   Given mirror-lens spacing, mirror semi-diameter,
|              initial ray height at lens' first surface, and
|              an initial object distance spacing, calculates
|              required object spacing (based on edge ray), new
|              ray height, and determines the resulting
|              spherical aberration between edge ray and
|              centre ray
|
--------------------------------------------------------------}
procedure CalculateSA (
        S  : extended;    { Mirror-lens spacing }
        H  : extended;    { Mirror semi-diameter }
    var Lo : extended;    { Initial object distance to update }
    var Y  : extended;    { Initial lens ray height to update }
    var SA : extended );  { Minimum Spherical Aberration given S }
var
    Li10, Li100, Y1, Y2 : extended;
begin

{ Determine object distance (= image distance) for 100% ray }
CalculateLo ( Y, S, Lo );

{ Adjust Y to keep Y2 close to mirror edge }
CalculateLi ( Y, Lo, S, Li100, Y1, Y2 );
Y := Y * H / Y2;

{ Recalculate Lo with new diameter }
CalculateLo ( Y, S, Lo );

{ SA: Calculate Li at 10% }
CalculateLi ( Y * 0.1, Lo, S, Li10, Y1, Y2 );

SA := Li100 - Li10;
end;

{--------------------------------------------------------------
|
|   Name:      Mainline
|   Purpose:   Performs Ross Null setup calculations
|
--------------------------------------------------------------}

var
    Lo, Li10, Li70, Li100, S, Y, Y1, Y2, SA, Tol, OPD : extended;
    Ch : char;
    St : String[ 20 ];
    RepeatInput : boolean;
    I : integer;
    AdjustGain, Adjust : extended;
    JustCheck, InputSet, GraphicsEnabled : boolean;
    lFL, lDiameter : extended;

begin
GraphicsOn := false;
GraphicsEnabled := false;
RepeatInput := false;
InputSet := false;
clrscr;

for I := 1 to ParamCount do
    begin
    if (Pos (ParamStr( I ), '/g' ) > 0) or
       (Pos (ParamStr( I ), '/G' ) > 0) then
        begin
        GraphicsEnabled := true;
        end
    else if not InputSet then
        begin
        InputSet := true;

        { Temporarily set input to a file }
        assign ( input, ParamStr( I ) );
        {$I-}
        reset ( input );
        {$I+}
        if IOResult = 0 then
            RepeatInput := true
        else
            begin
            assign ( input, 'con' );
            reset ( input );
            end;
        end
    else
        begin
        InputSet := true;

        { set output to a file }
        assign ( output, ParamStr( I ) );
        {$I-}
        rewrite ( output );
        {$I+}
        if IOResult <> 0 then
            begin
            assign ( output, 'con' );
            rewrite ( output );
            end;
        end;
    end;

writeln ( 'Ross Null Test Setup Calculations V1.1' );
writeln ( 'Copyright (c) 1990, Doug George' );
writeln ( '66 Shetland Way, Kanata, Ontario, Canada, K2M 1S4' );
writeln ( 'This shareware program may be freely distributed and used' );
writeln;
writeln ( 'All Dimensions in Inches' );
writeln;
repeat
    writeln ( 'Mirror Parameters' );
    writeln;
    write ( 'Diameter                       -> ' );
    readln ( mSemiDiameter );
    if RepeatInput then writeln ( mSemiDiameter:11:6 );
    mSemiDiameter := mSemiDiameter * 0.5;

    write ( 'Focal Length                   -> ' );
    readln ( mRadius );
    if RepeatInput then writeln ( mRadius:11:6 );
    mRadius := -abs ( 2 * mRadius );

    writeln ( 'Conic Const. = 1 - Eccent. ** 2' );
    writeln ( 'Parabola = 0, Sphere = 1' );
    write ( 'Conic Constant                 -> ' );
    readln ( mConicConst );
    if RepeatInput then writeln ( mConicConst:11:6 );

    writeln;
    writeln ( 'Lens Parameters' );
    writeln;
    writeln ( 'Sign Convention:' );
    writeln ( 'With mirror on right, ) = -ve, ( = +ve' );
    writeln;
    write ( 'Image Side Radius              -> ' );
    readln ( lRadiusI );
    if RepeatInput then
        begin
        if abs ( lRadiusI ) > 1000 then
            writeln ( '   Infinite' )
        else
            writeln ( lRadiusI:11:6 );
        end;

    write ( 'Mirror Side Radius             -> ' );
    readln ( lRadiusM );
    if RepeatInput then
        begin
        if abs ( lRadiusM ) > 1000 then
            writeln ( '   Infinite' )
        else
            writeln ( lRadiusM:11:6 );
        end;

    write ( 'Thickness                      -> ' );
    readln ( lThick );
    if RepeatInput then writeln ( lThick:11:6 );

    write ( 'Diameter                       -> ' );
    readln ( lDiameter );
    if RepeatInput then writeln ( lDiameter:11:6 );

    write ( 'Refractive Index               -> ' );
    readln ( lIndex );
    if RepeatInput then writeln ( lIndex:11:6 );
    writeln;

    { Calculate Lens Focal Length }
    lFL := 1.0 / ( ( lIndex - 1 ) * ( -1.0 / lRadiusM + 1.0 / lRadiusI ) );

    writeln ( 'Enter Lens To Mirror Distance' );
    write ( 'Or Enter 0 to Calculate        -> ' );
    readln ( S );
    if RepeatInput then writeln ( S:11:6 );
    JustCheck := S <> 0;
    writeln;

    { Reset input to console }
    close ( input );
    assign ( input, 'con' );
    reset ( input );

    { Guess initial Lo based on 1/2 lens focal length }
    Lo := lFL * 0.5;

    { Guess S at mirror radius minus lens focal length }
    if not JustCheck then
        S := -mRadius - lFL;

    { Guess Y from simple geometry }
    Y := (mRadius + S) / mRadius * mSemiDiameter;

    { Pick useable adjustment gain factor using mirror and lens F/ratio }
    AdjustGain := Power (-mRadius / mSemiDiameter / 4, 1.5 ) * lFL / lDiameter;

    if GraphicsEnabled then
        begin
        GraphicsOn := true;
        InitGraphics;
        XScale := -Round ( GetMaxX / mRadius );
        YScale := -Round ( GetMaxY / (mSemiDiameter * 3.1) );
        end;

    if GraphicsOn then SetViewPort ( 0, 0, GetMaxX, GetMaxY - 100, ClipOn );

    repeat
        if GraphicsOn then SetViewPort ( 0, 0, GetMaxX, GetMaxY - 100, ClipOn );

        CalculateSA ( S, mSemiDiameter, Lo, Y, SA );

        if GraphicsOn then
            begin
            SetViewPort ( 0, GetMaxY - 100, GetMaxX, GetMaxY, ClipOn );
            ClearViewPort;
            SetColor ( White );
            str ( Lo:11:6, St );
            OutTextXY ( 0, 10, 'Lo   = ' + St );
            str ( S:11:6, St );
            OutTextXY ( 0, 20, 'S    = ' + St );
            str ( SA:11:6, St );
            OutTextXY ( 0, 30, 'SA   = ' + St );
            ToggleScreen;
            ClearViewPort;
            str ( Lo:11:6, St );
            OutTextXY ( 0, 10, 'Lo   = ' + St );
            str ( S:11:6, St );
            OutTextXY ( 0, 20, 'S    = ' + St );
            str ( SA:11:6, St );
            OutTextXY ( 0, 30, 'SA   = ' + St );
            ToggleScreen;
            end
        else
            begin
            writeln ( 'Object-Lens = ', Lo:11:6,
                    '  Lens-Mirror = ', S:11:6,
                    '  Marginal SA = ', SA:11:6 );
            GotoXY ( 1, WhereY - 1 );
            end;

        if (abs ( SA ) > SAResidual) and not JustCheck then
            begin
            Adjust := SA * AdjustGain;
            if Adjust > 1 then
                Adjust := 1
            else if Adjust < -1 then
                Adjust := -1;
            S := S + Adjust;
            end;

    until (abs ( SA ) < SAResidual) or keypressed or JustCheck;

    if GraphicsOn then
        begin
        GraphicsOn := false;
        RestoreCRTMode;
        end;

    { Calculate a few output parameters }
    CalculateLi ( Y * 0.1, Lo, S, Li10, Y1, Y2 );  { To get Li10 }
    CalculateLi ( Y * 0.7, Lo, S, Li70, Y1, Y2 );  { To get Li70 }
    CalculateLi ( Y, Lo, S, Li100, Y1, Y2 );       { To get Y1 }
    Tol := 0.008 / 25.4 * sqr ( Lo / (2 * Y) );

    OPD := abs (Li70 - Li10);
    if OPD < abs ( Li100 - Li10 ) then
        OPD := abs ( Li100 - Li10 );
    if OPD < abs ( Li100 - Li70 ) then
        OPD := abs ( Li100 - Li70 );
    OPD := OPD / (4 * Tol);

    writeln;
    writeln;
    writeln ( 'Results:' );
    writeln;
    writeln ( 'Object to Lens Distance         = ', Lo:11:6 );
    writeln ( 'Lens to Mirror Distance         = ', S:11:6 );
    writeln ( 'Working Diameter of Lens        = ', Y1 * 2:11:6 );
    writeln ( '70% Zone Spherical Aberration   = ', Li70 - Li10:11:6 );
    writeln ( 'Marginal Spherical Aberration   = ', Li100 - Li10:11:6 );
    writeln ( 'Quarter-wave Tolerence          = ', Tol:11:6 );
    writeln ( 'Maximum Optical Path Difference = ', OPD:11:6 );

    if GraphicsEnabled then Halt;

    { Reset output to console }
    close ( output );
    assign ( output, 'con' );
    rewrite ( output );

    writeln;
    write ( 'Another Setup? (Y/N) ' );
    readln ( Ch );
    RepeatInput := false;
until (Ch <> 'Y') and (Ch <> 'y');
end.
